
package sml.geoUtils;

import sml.geoUtils.geometry.IPointGeo;
import sml.geoUtils.geometry.IPolyline;
import sml.geoUtils.geometry.PointGeo;
import sml.geoUtils.geometry.Segment;
import sml.gps.datamodel.GPSPoint;

public class Calculations {
    public static Double EARTH_RADIOUS = 6371010.0;
    public static Double EPS_LENGTH    = 0.01;
    public static int NUMBEROFPOINTS = 1000;
    //static GreatCircleDistanceCalculator distanceCalculator;
    static IDistanceCalculator distanceCalculator= new GreatCircleDistanceCalculator();
    public Calculations(){
        distanceCalculator = new GreatCircleDistanceCalculator();
    }
    
    public static Double getBearing(IPointGeo pt1, IPointGeo pt2){
        Double dLon = Calculations.toRadians(pt1.getLongitude()-pt2.getLongitude());
        Double dLat = Calculations.toRadians(pt1.getLatitude()-pt2.getLatitude());
        
        Double y = Math.sin(dLon) * Math.cos(Calculations.toRadians(pt2.getLatitude()));
        Double x = Math.cos(Calculations.toRadians(pt1.getLatitude())) * Math.sin(Calculations.toRadians(pt2.getLatitude()))
                - Math.sin(Calculations.toRadians(pt1.getLatitude())) * Math.cos(Calculations.toRadians(pt2.getLatitude())) * Math.cos(dLon);
        return Calculations.toDegrees(Math.atan2(y, x));
    }
    
    public static Double getBearing(Segment<IPointGeo> segment){
        return Calculations.getBearing(segment.getStartPoint(), segment.getEndPoint());
    };
    
    public static Double getDistance2D(IPointGeo pt1, IPointGeo pt2){
        //System.out.println("pt1=<"+pt1.getLatitude()+","+pt1.getLongitude()+">");
        //System.out.println("pt1=<"+pt2.getLatitude()+","+pt2.getLongitude()+">");
        if( pt1 == null ){ System.err.println("Calculations.getDistance:::pt1=null"); System.exit(1);}
        if( pt2 == null ){ System.err.println("Calculations.getDistance:::pt2=null"); System.exit(1);}
        return distanceCalculator.calculate2D(pt1,pt2);
    }
    
    public static Double getDistance2D(IPointGeo point, Segment<IPointGeo> segment){
          IPointGeo projectedPoint = Topology.projectPoint(point, segment);
          return distanceCalculator.calculate2D(point, projectedPoint);
    }
    
    public static Double getDistance2D(IPointGeo point, IPolyline<IPointGeo> line){
          Double minDistance = Double.POSITIVE_INFINITY;
          for(Segment segment : line.getSegments()){
             IPointGeo projectedPoint = Topology.projectPoint(point,segment);
             minDistance = Math.min(minDistance, distanceCalculator.calculate2D(point,projectedPoint));
          }
          return minDistance;
    }
    
    public static Double getDistance2D(IPointGeo point, IPolyline<IPointGeo> line, int closestSegmentIndex){
        Double minDistance = Double.POSITIVE_INFINITY;
        for(int i=0;i<line.getSegments().size();i++){
           IPointGeo projectedPoint = Topology.projectPoint(point,line.getSegments().get(i));
           Double distance = distanceCalculator.calculate2D(point, projectedPoint);
           if ( distance < minDistance ){
               minDistance = distance;
               closestSegmentIndex = i;
           }
        }
        return minDistance;
    }
    public static Double getPathLength(IPointGeo from, IPointGeo to, Segment<IPointGeo> path){
        return Calculations.getDistance2D(from, to);
    }
    
    public static Double getPathLength(IPointGeo from, IPointGeo to, IPolyline<IPointGeo> path){
       int pointFound = 0;
       int pointFoundLast = 0;
       boolean fromFound = false; 
       boolean toFound   = false;
       Double distance = 0.0;
       for(Segment segment : path.getSegments()){
          pointFoundLast = pointFound;
          IPointGeo[] points = new IPointGeo[] {segment.getStartPoint(),segment.getEndPoint()};
          if (!fromFound && (Calculations.getDistance2D(from, segment) < EPS_LENGTH)){
             fromFound = true;
             points[pointFound++] = from;
          }
          if(!toFound    && (Calculations.getDistance2D(to,   segment) < EPS_LENGTH)){
             toFound = true;
             points[pointFound++] = to;
          }
          if ( pointFound > 0 ){
             if (pointFound == pointFoundLast)  distance = distance + segment.getLength();
             else                               distance = distance + Calculations.getDistance2D(points[0],points[1]);
             if (pointFound == 2) return distance;
          
          }
       }
       return distance;
    }
    public static Double getLength(IPolyline<IPointGeo> line){
        Double length = 0.0;
        for(Segment segment : line.getSegments()){
           length = length + segment.getLength();
        }
        return length;
    }
    public static PointGeo Averaging(Iterable<IPointGeo> points){
       int count = 0;
       Double lat = 0.0;
       Double lon = 0.0;
       for(IPointGeo point : points){
           lat = lat + point.getLatitude();
           lon = lon + point.getLongitude();
           count = count + 1;
       }
       return new PointGeo(lat/count, lon/count);
    }
    public static Double toRadians(Double angle){
       return angle * Math.PI / 180.0;
    }
    public static double toDegrees(Double angle){
       return angle * 180.0 / Math.PI;
    }
    
    public static GPSPoint greatCirclePoint(GPSPoint pnt1, GPSPoint pnt2, Double f){
        /*
        https://math.stackexchange.com/questions/383711/parametric-equation-for-great-circle
            d = acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(lon1 - lon2))
            A = sin((1 - f) * d) / sin(d)
            B = sin(f * d) / sin(d)
            x = A * cos(lat1) * cos(lon1) + B * cos(lat2) * cos(lon2)
            y = A * cos(lat1) * sin(lon1) + B * cos(lat2) * sin(lon2)
            z = A * sin(lat1) + B * sin(lat2)

            lat_f = atan2(z, sqrt(x^2 + y^2))
            lon_f = atan2(y,x)
            } 
        */  
        
        Double lat1 = Math.toRadians(pnt1.getLatitude());
        Double lon1 = Math.toRadians(pnt1.getLongitude());
        Double lat2 = Math.toRadians(pnt2.getLatitude());
        Double lon2 = Math.toRadians(pnt2.getLongitude());
        
        Double d = Math.acos(Math.sin(lat1) * Math.sin(lat2) + Math.cos(lat1) * Math.cos(lat2) * Math.cos(lon1-lon2));
        Double A = Math.sin((1.0-f) * d ) / Math.sin(d);
        Double B = Math.sin(f*d) / Math.sin(d);
        Double x = A * Math.cos(lat1) * Math.cos(lon1) + B * Math.cos(lat2) * Math.cos(lon2);
        Double y = A * Math.cos(lat1) * Math.sin(lon1) + B * Math.cos(lat2) * Math.sin(lon2);
        Double z = A * Math.sin(lat1) + B * Math.sin(lat2);
        
        Double lat_f = Math.toDegrees(Math.atan2(z, Math.sqrt(x*x+y*y)));
        Double lon_f = Math.toDegrees(Math.atan2(y,x));
        
        GPSPoint pnt = new GPSPoint(lat_f,lon_f);
        return pnt;
    }
    
    public static Double distanceToGreatCircle(GPSPoint pnt, GPSPoint pnt1, GPSPoint pnt2){
        Double minDistance = Double.MAX_VALUE;
        int numPoints = NUMBEROFPOINTS;
        for(int i=0;i<numPoints;i++){
            Double f = (i*1.0) / ((numPoints-1)*1.0);
            GPSPoint gcPnt = greatCirclePoint(pnt1, pnt2, f);
            Double distance = Calculations.getDistance2D(pnt, gcPnt);
            if ( distance < minDistance ){
                minDistance = distance;
            }
        }
        return minDistance;
    }
}
